// SPDX-License-Identifier: MIT
pragma solidity ^0.8.0;

import "../lib/MetaKeepLambdaStructs.sol";
import "../lib/MetaKeepLambdaConstants.sol";
import "./MetaKeepLambdaEIP712Base.sol";

abstract contract MetaKeepLambdaBase is
    MetaKeepLambdaConstants,
    MetaKeepLambdaEIP712Base
{
    mapping(address => mapping(uint256 => bool)) private nonceStatus;

    //slither-disable-next-line locked-ether,assembly,low-level-calls
    function executeMetaKeepLambdaTransaction(
        address userAddress,
        uint256 nonce,
        uint256 validFrom,
        uint256 validUpto,
        bytes memory functionSignature,
        bytes32 sigR,
        bytes32 sigS,
        uint8 sigV
    ) external payable returns (bool) {
        _requireValidAuthorization(userAddress, nonce, validFrom, validUpto);

        MetaKeepLambdaStructs.MetaKeepLambdaTransaction
            memory metaTx = MetaKeepLambdaStructs.MetaKeepLambdaTransaction({
                nonce: nonce,
                validFrom: validFrom,
                validUpto: validUpto,
                from: userAddress,
                functionSignature: functionSignature
            });

        bytes32 data = keccak256(
            abi.encode(
                METAKEEP_LAMBDA_TRANSACTION_TYPEHASH,
                metaTx.nonce,
                metaTx.validFrom,
                metaTx.validUpto,
                metaTx.from,
                keccak256(metaTx.functionSignature)
            )
        );
        require(
            _verify(userAddress, data, sigR, sigS, sigV),
            "MetaKeepLambda: Sig not match"
        );
        // disable nonce for user (to avoid re-use)
        nonceStatus[userAddress][nonce] = true;

        // Append userAddress and relayer address at the end to extract it from calling context
        (bool success, bytes memory returnData) = address(this).call(
            abi.encodePacked(functionSignature, userAddress)
        );
        if (!success) {
            assembly {
                revert(add(returnData, 32), returnData)
            }
        }

        return success;
    }

    function metaKeepLambdaNonceState(address userAddress, uint256 nonce)
        external
        view
        returns (bool)
    {
        return nonceStatus[userAddress][nonce];
    }

    function cancelMetaKeepLambdaNonce(
        address userAddress,
        uint256 nonce,
        bytes32 sigR,
        bytes32 sigS,
        uint8 sigV
    ) external {
        _requireUnusedNonce(userAddress, nonce);

        bytes32 data = keccak256(
            abi.encode(
                METAKEEP_LAMBDA_CANCEL_NONCE_TYPEHASH,
                nonce,
                userAddress
            )
        );
        require(
            _verify(userAddress, data, sigR, sigS, sigV),
            "MetaKeepLambda: Sig not match"
        );

        // Set nonce to used.
        nonceStatus[userAddress][nonce] = true;
    }

    //slither-disable-next-line timestamp
    function _requireValidAuthorization(
        address userAddress,
        uint256 nonce,
        uint256 validFrom,
        uint256 validUpto
    ) private view {
        require(
            block.timestamp > validFrom,
            "MetaKeepLambda: Authorization is not yet given"
        );
        require(
            block.timestamp < validUpto,
            "MetaKeepLambda: Authorization has expired"
        );
        _requireUnusedNonce(userAddress, nonce);
    }

    function _requireUnusedNonce(address userAddress, uint256 nonce)
        private
        view
    {
        require(!nonceStatus[userAddress][nonce], "MetaKeepLambda: Nonce Used");
    }

    function _verify(
        address signer,
        bytes32 data,
        bytes32 sigR,
        bytes32 sigS,
        uint8 sigV
    ) private view returns (bool) {
        require(signer != address(0), "MetaKeepLambda: INVALID_SIGNER");
        return signer == ecrecover(toTypedMessageHash(data), sigV, sigR, sigS);
    }
}
